# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable

import awkward as ak
from awkward._v2.index import Index
from awkward._v2._slicing import NestedIndexError
from awkward._v2.contents.content import Content
from awkward._v2.forms.unionform import UnionForm

np = ak.nplike.NumpyMetadata.instance()


class UnionArray(Content):
    def __init__(self, tags, index, contents, identifier=None, parameters=None):
        if not (isinstance(tags, Index) and tags.dtype == np.dtype(np.int8)):
            raise TypeError(
                "{0} 'tags' must be an Index with dtype=int8, not {1}".format(
                    type(self).__name__, repr(tags)
                )
            )
        if not isinstance(index, Index) and index.dtype in (
            np.dtype(np.int32),
            np.dtype(np.uint32),
            np.dtype(np.int64),
        ):
            raise TypeError(
                "{0} 'index' must be an Index with dtype in (int32, uint32, int64), "
                "not {1}".format(type(self).__name__, repr(index))
            )

        if not isinstance(contents, Iterable):
            raise TypeError(
                "{0} 'contents' must be iterable, not {1}".format(
                    type(self).__name__, repr(contents)
                )
            )
        if not isinstance(contents, list):
            contents = list(contents)

        for content in contents:
            if not isinstance(content, Content):
                raise TypeError(
                    "{0} all 'contents' must be Content subclasses, not {1}".format(
                        type(self).__name__, repr(content)
                    )
                )

        if not len(tags) <= len(index):
            raise ValueError(
                "{0} len(tags) ({1}) must be <= len(index) ({2})".format(
                    type(self).__name__, len(tags), len(index)
                )
            )

        self._tags = tags
        self._index = index
        self._contents = contents
        self._init(identifier, parameters)

    @property
    def tags(self):
        return self._tags

    @property
    def index(self):
        return self._index

    def content(self, index):
        return self._contents[index]

    @property
    def contents(self):
        return self._contents

    @property
    def nplike(self):
        return self._tags.nplike

    @property
    def nonvirtual_nplike(self):
        return self._tags.nplike

    Form = UnionForm

    @property
    def form(self):
        return self.Form(
            self._tags.form,
            self._index.form,
            [x.form for x in self._contents],
            has_identifier=self._identifier is not None,
            parameters=self._parameters,
            form_key=None,
        )

    @property
    def typetracer(self):
        tt = ak._v2._typetracer.TypeTracer.instance()
        return UnionArray(
            ak._v2.index.Index(self._tags.to(tt)),
            ak._v2.index.Index(self._index.to(tt)),
            [x.typetracer for x in self._contents],
            self._typetracer_identifier(),
            self._parameters,
        )

    def __len__(self):
        return len(self._tags)

    def __repr__(self):
        return self._repr("", "", "")

    def _repr(self, indent, pre, post):
        out = [indent, pre, "<UnionArray len="]
        out.append(repr(str(len(self))))
        out.append(">")
        out.extend(self._repr_extra(indent + "    "))
        out.append("\n")
        out.append(self._tags._repr(indent + "    ", "<tags>", "</tags>\n"))
        out.append(self._index._repr(indent + "    ", "<index>", "</index>\n"))

        for i, x in enumerate(self._contents):
            out.append("{0}    <content index={1}>\n".format(indent, repr(str(i))))
            out.append(x._repr(indent + "        ", "", "\n"))
            out.append("{0}    </content>\n".format(indent))

        out.append(indent + "</UnionArray>")
        out.append(post)
        return "".join(out)

    def _getitem_nothing(self):
        return self._getitem_range(slice(0, 0))

    def _getitem_at(self, where):
        if where < 0:
            where += len(self)
        if not (0 <= where < len(self)) and self.nplike.known_shape:
            raise NestedIndexError(self, where)
        tag, index = self._tags[where], self._index[where]
        return self._contents[tag]._getitem_at(index)

    def _getitem_range(self, where):
        start, stop, step = where.indices(len(self))
        assert step == 1
        return UnionArray(
            self._tags[start:stop],
            self._index[start:stop],
            self._contents,
            self._range_identifier(start, stop),
            self._parameters,
        )

    def _getitem_field(self, where, only_fields=()):
        return UnionArray(
            self._tags,
            self._index,
            [x._getitem_field(where, only_fields) for x in self._contents],
            self._field_identifier(where),
            None,
        )

    def _getitem_fields(self, where, only_fields=()):
        return UnionArray(
            self._tags,
            self._index,
            [x._getitem_fields(where, only_fields) for x in self._contents],
            self._fields_identifer(where),
            None,
        )

    def _carry(self, carry, allow_lazy, exception):
        assert isinstance(carry, ak._v2.index.Index)

        try:
            nexttags = self._tags[carry.data]
            nextindex = self._index[: len(self._tags)][carry.data]
        except IndexError as err:
            if issubclass(exception, NestedIndexError):
                raise exception(self, carry.data, str(err))
            else:
                raise exception(str(err))

        return UnionArray(
            nexttags,
            nextindex,
            self._contents,
            self._carry_identifier(carry, exception),
            self._parameters,
        )

    def _project(self, index):
        nplike = self.nplike
        lentags = len(self._tags)
        assert len(self._index) == lentags
        lenout = ak._v2.index.Index64.empty(1, nplike)
        tmpcarry = ak._v2.index.Index64.empty(lentags, nplike)
        self._handle_error(
            nplike[
                "awkward_UnionArray_project",
                lenout.dtype.type,
                tmpcarry.dtype.type,
                self._tags.dtype.type,
                self._index.dtype.type,
            ](
                lenout.to(nplike),
                tmpcarry.to(nplike),
                self._tags.to(nplike),
                self._index.to(nplike),
                lentags,
                index,
            )
        )
        nextcarry = ak._v2.index.Index64(tmpcarry.data[: lenout[0]], nplike)
        return self._contents[index]._carry(nextcarry, False, NestedIndexError)

    def _regular_index(self, tags):
        nplike = self.nplike
        lentags = len(tags)
        size = ak._v2.index.Index64.empty(1, nplike)
        self._handle_error(
            nplike[
                "awkward_UnionArray_regular_index_getsize",
                size.dtype.type,
                tags.dtype.type,
            ](
                size.to(nplike),
                tags.to(nplike),
                lentags,
            )
        )
        current = ak._v2.index.Index64.empty(size[0], nplike)
        outindex = ak._v2.index.Index64.empty(lentags, nplike)
        self._handle_error(
            nplike[
                "awkward_UnionArray_regular_index",
                outindex.dtype.type,
                current.dtype.type,
                tags.dtype.type,
            ](
                outindex.to(nplike),
                current.to(nplike),
                size[0],
                tags.to(nplike),
                lentags,
            )
        )
        return outindex

    def _getitem_next_jagged_generic(self, slicestarts, slicestops, slicecontent, tail):
        simplified = self._simplify_uniontype()
        if (
            simplified.index.dtype == np.dtype(np.int32)
            or simplified.index.dtype == np.dtype(np.uint32)
            or simplified.index.dtype == np.dtype(np.int64)
        ):
            raise NestedIndexError(
                self,
                ak._v2.contents.ListArray(slicestarts, slicestops, slicecontent),
                "cannot apply jagged slices to irreducible union arrays",
            )
        return simplified._getitem_next_jagged(
            slicestarts, slicestops, slicecontent, tail
        )

    def _getitem_next_jagged(self, slicestarts, slicestops, slicecontent, tail):
        return self._getitem_next_jagged_generic(
            slicestarts, slicestops, slicecontent, tail
        )

    def _getitem_next(self, head, tail, advanced):
        if head == ():
            return self

        elif isinstance(
            head, (int, slice, ak._v2.index.Index64, ak._v2.contents.ListOffsetArray)
        ):
            outcontents = []
            for i in range(len(self._contents)):
                projection = self._project(i)
                outcontents.append(projection._getitem_next(head, tail, advanced))
            outindex = self._regular_index(self._tags)

            out = UnionArray(
                self._tags,
                outindex,
                outcontents,
                self._identifier,
                self._parameters,
            )
            return out._simplify_uniontype()

        elif ak._util.isstr(head):
            return self._getitem_next_field(head, tail, advanced)

        elif isinstance(head, list):
            return self._getitem_next_fields(head, tail, advanced)

        elif head is np.newaxis:
            return self._getitem_next_newaxis(tail, advanced)

        elif head is Ellipsis:
            return self._getitem_next_ellipsis(tail, advanced)

        elif isinstance(head, ak._v2.contents.IndexedOptionArray):
            return self._getitem_next_missing(head, tail, advanced)

        else:
            raise AssertionError(repr(head))

    def _localindex(self, axis, depth):
        posaxis = self._axis_wrap_if_negative(axis)
        if posaxis == depth:
            return self._localindex_axis0()
        else:
            contents = []
            for content in self._contents:
                contents.append(content._localindex(posaxis, depth))
            return UnionArray(
                self._tags, self._index, contents, self._identifier, self._parameters
            )

    def _combinations(self, n, replacement, recordlookup, parameters, axis, depth):
        posaxis = self._axis_wrap_if_negative(axis)
        if posaxis == depth:
            return self._combinations_axis0(n, replacement, recordlookup, parameters)
        else:
            contents = []
            for content in self._contents:
                contents.append(
                    content._combinations(
                        n, replacement, recordlookup, parameters, posaxis, depth
                    )
                )
            return ak._v2.unionarray.UnionArray(
                self._tags, self._index, contents, self._identifier, self._parameters
            )

    def _argsort_next(
        self,
        negaxis,
        starts,
        shifts,
        parents,
        outlength,
        ascending,
        stable,
        kind,
        order,
    ):
        simplified = self._simplify_uniontype()
        if isinstance(simplified, ak._v2.contents.UnionArray):
            raise ValueError("cannot argsort an irreducible UnionArray")

        return simplified._argsort_next(
            negaxis, starts, shifts, parents, outlength, ascending, stable, kind, order
        )

    def _sort_next(
        self, negaxis, starts, parents, outlength, ascending, stable, kind, order
    ):
        simplified = self._simplify_uniontype()
        if isinstance(simplified, ak._v2.contents.UnionArray):
            raise ValueError("cannot sort an irreducible UnionArray")

        return simplified._sort_next(
            negaxis, starts, parents, outlength, ascending, stable, kind, order
        )

    def _reduce_next(
        self,
        reducer,
        negaxis,
        starts,
        shifts,
        parents,
        outlength,
        mask,
        keepdims,
    ):
        simplified = self._simplify_uniontype()
        if isinstance(simplified, UnionArray):
            raise ValueError(
                "cannot call ak.{0} on an irreducible UnionArray".format(reducer.name)
            )

        return simplified._reduce_next(
            reducer,
            negaxis,
            starts,
            shifts,
            parents,
            outlength,
            mask,
            keepdims,
        )

    def _validityerror(self, path):
        for i in range(len(self.contents)):
            if isinstance(self.contents[i], ak._v2.contents.unionarray.UnionArray):
                return "{0} contains {1}, the operation that made it might have forgotten to call 'simplify_uniontype'".format(
                    type(self), type(self.contents[i])
                )
            if len(self.index) < len(self.tags):
                return 'at {0} ("{1}"): len(index) < len(tags)'.format(path, type(self))
            lencontents = self.nplike.empty(len(self.contents), dtype=np.int64)
            for i in range(len(self.contents)):
                lencontents[i] = len(self.contents[i])
            error = self.nplike[
                "awkward_UnionArray_validity",
                self.tags.dtype.type,
                self.index.dtype.type,
                np.int64,
            ](
                self.tags.to(self.nplike),
                self.index.to(self.nplike),
                len(self.tags),
                len(self.contents),
                lencontents,
            )
            if error.str is not None:
                if error.filename is None:
                    filename = ""
                else:
                    filename = " (in compiled code: " + error.filename.decode(
                        errors="surrogateescape"
                    ).lstrip("\n").lstrip("(")
                message = error.str.decode(errors="surrogateescape")
                return 'at {0} ("{1}"): {2} at i={3}{4}'.format(
                    path, type(self), message, error.id, filename
                )
            for i in range(len(self.contents)):
                sub = self.contents[i].validityerror(path + ".content({0})".format(i))
                if sub != "":
                    return sub
            return ""



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_singletons():
    array = ak.Array([1.1, 2.2, None, 3.3, None, None, 4.4, 5.5])
    assert ak.to_list(ak.singletons(array)) == [
        [1.1],
        [2.2],
        [],
        [3.3],
        [],
        [],
        [4.4],
        [5.5],
    ]

    assert ak.to_list(
        ak.singletons(ak.Array([1.1, 2.2, None, 3.3, None, None, 4.4, 5.5]))
    ) == [[1.1], [2.2], [], [3.3], [], [], [4.4], [5.5]]
    assert ak.to_list(
        ak.singletons(ak.Array([[1.1, 2.2, None], [3.3, None], [None], [4.4, 5.5]]))
    ) == [[[1.1], [2.2], []], [[3.3], []], [[]], [[4.4], [5.5]]]
    assert ak.to_list(
        ak.singletons(
            ak.Array([[[1.1, 2.2, None]], [[3.3, None]], [[None]], [[4.4, 5.5]]])
        )
    ) == [[[[1.1], [2.2], []]], [[[3.3], []]], [[[]]], [[[4.4], [5.5]]]]


def test_firsts():
    assert ak.to_list(
        ak.firsts(
            ak.singletons(ak.Array([1.1, 2.2, None, 3.3, None, None, 4.4, 5.5])), axis=1
        )
    ) == [1.1, 2.2, None, 3.3, None, None, 4.4, 5.5]
    assert (
        ak.to_list(
            ak.firsts(
                ak.singletons(
                    ak.Array([[1.1, 2.2, None], [3.3, None], [None], [4.4, 5.5]])
                ),
                axis=2,
            )
        )
        == [[1.1, 2.2, None], [3.3, None], [None], [4.4, 5.5]]
    )
    assert (
        ak.to_list(
            ak.firsts(
                ak.singletons(
                    ak.Array(
                        [[[1.1, 2.2, None]], [[3.3, None]], [[None]], [[4.4, 5.5]]]
                    )
                ),
                axis=3,
            )
        )
        == [[[1.1, 2.2, None]], [[3.3, None]], [[None]], [[4.4, 5.5]]]
    )


def test_allow_missing():
    array = ak.Array([1.1, 2.2, None, 3.3, None, None, 4.4, 5.5])
    ak.to_numpy(array)
    with pytest.raises(ValueError):
        ak.to_numpy(array, allow_missing=False)


def test_flatten0():
    array = ak.Array([1.1, 2.2, None, 3.3, None, None, 4.4, 5.5])
    assert ak.to_list(ak.flatten(array, axis=0)) == [1.1, 2.2, 3.3, 4.4, 5.5]

    content0 = ak.from_iter(
        [1.1, 2.2, None, 3.3, None, None, 4.4, 5.5], highlevel=False
    )
    content1 = ak.from_iter(["one", None, "two", None, "three"], highlevel=False)
    array = ak.Array(
        ak.layout.UnionArray8_64(
            ak.layout.Index8(
                np.array([0, 1, 0, 1, 0, 1, 0, 1, 0, 0, 0, 1, 0], dtype=np.int8)
            ),
            ak.layout.Index64(
                np.array([0, 0, 1, 1, 2, 2, 3, 3, 4, 5, 6, 4, 7], dtype=np.int64)
            ),
            [content0, content1],
        )
    )
    assert ak.to_list(array) == [
        1.1,
        "one",
        2.2,
        None,
        None,
        "two",
        3.3,
        None,
        None,
        None,
        4.4,
        "three",
        5.5,
    ]
    assert ak.to_list(ak.flatten(array, axis=0)) == [
        1.1,
        "one",
        2.2,
        "two",
        3.3,
        4.4,
        "three",
        5.5,
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    array = ak.repartition([{"x": x, "y": x * 10} for x in range(10)], 2)
    assert ak.to_list(ak.num(array.layout, axis=0)) == {"x": 10, "y": 10}



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_fromnumpy():
    a = np.arange(2 * 3 * 5).reshape((2, 3, 5))
    b = ak.from_numpy(a)
    assert ak.to_list(a) == ak.to_list(b)


def test_highlevel():
    a = ak.Array(
        [[1.1, 2.2, 3.3], [], [4.4, 5.5], [6.6], [7.7, 8.8, 9.9]], check_valid=True
    )
    assert (
        repr(a)
        == "<Array [[1.1, 2.2, 3.3], ... [7.7, 8.8, 9.9]] type='5 * var * float64'>"
    )
    assert str(a) == "[[1.1, 2.2, 3.3], [], [4.4, 5.5], [6.6], [7.7, 8.8, 9.9]]"

    b = ak.Array(np.arange(100, dtype=np.int32), check_valid=True)
    assert (
        repr(b) == "<Array [0, 1, 2, 3, 4, ... 95, 96, 97, 98, 99] type='100 * int32'>"
    )
    assert (
        str(b)
        == "[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, ... 89, 90, 91, 92, 93, 94, 95, 96, 97, 98, 99]"
    )

    c = ak.Array(
        '[{"one": 3.14, "two": [1.1, 2.2]}, {"one": 99.9, "two": [-3.1415926]}]',
        check_valid=True,
    )
    assert (
        repr(c)
        == '<Array [{one: 3.14, two: [1.1, ... -3.14]}] type=\'2 * {"one": float64, "two": va...\'>'
    )
    assert str(c) == "[{one: 3.14, two: [1.1, 2.2]}, {one: 99.9, two: [-3.14]}]"


class Dummy(ak.highlevel.Array):
    pass


def test_string1():
    a = ak.Array(
        np.array([ord(x) for x in "hey there"], dtype=np.uint8), check_valid=True
    )
    a.__class__ = ak.behaviors.string.ByteBehavior
    assert str(a) == str(b"hey there")
    assert repr(a) == repr(b"hey there")


def test_string2():
    content = ak.layout.NumpyArray(
        np.array([ord(x) for x in "heythere"], dtype=np.uint8)
    )
    listoffsetarray = ak.layout.ListOffsetArray64(
        ak.layout.Index64(np.array([0, 3, 3, 8])), content
    )
    a = ak.Array(listoffsetarray, check_valid=True)

    assert isinstance(a, ak.Array)
    assert not isinstance(a, ak.behaviors.string.StringBehavior)
    assert ak.to_list(a) == [[104, 101, 121], [], [116, 104, 101, 114, 101]]

    assert repr(ak.type(a)) == "3 * var * uint8"
    assert repr(ak.type(a[0])) == "3 * uint8"
    assert repr(ak.type(a[1])) == "0 * uint8"
    assert repr(ak.type(a[2])) == "5 * uint8"

    assert (
        repr(a)
        == "<Array [[104, 101, 121], ... 101, 114, 101]] type='3 * var * uint8'>"
    )
    assert str(a) == "[[104, 101, 121], [], [116, 104, 101, 114, 101]]"
    assert repr(a[0]) == "<Array [104, 101, 121] type='3 * uint8'>"
    assert repr(a[1]) == "<Array [] type='0 * uint8'>"
    assert repr(a[2]) == "<Array [116, 104, 101, 114, 101] type='5 * uint8'>"

    content = ak.layout.NumpyArray(
        np.array([ord(x) for x in "heythere"], dtype=np.uint8),
        parameters={"__array__": "char", "encoding": "utf-8"},
    )
    listoffsetarray = ak.layout.ListOffsetArray64(
        ak.layout.Index64(np.array([0, 3, 3, 8])),
        content,
        parameters={"__array__": "string"},
    )
    a = ak.Array(listoffsetarray, check_valid=True)

    a = ak.Array(listoffsetarray, check_valid=True)
    assert isinstance(a, ak.Array)
    assert ak.to_list(a) == ["hey", "", "there"]

    if ak._util.py27:
        assert str(a) == "[u'hey', u'', u'there']"
        assert repr(a[0]) == "u'hey'"
        assert repr(a[1]) == "u''"
        assert repr(a[2]) == "u'there'"
    else:
        assert str(a) == "['hey', '', 'there']"
        assert repr(a[0]) == "'hey'"
        assert repr(a[1]) == "''"
        assert repr(a[2]) == "'there'"



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

from awkward._v2.tmp_for_testing import v1_to_v2, v1_to_v2_index

pytestmark = pytest.mark.skipif(
    ak._util.py27, reason="No Python 2.7 support in Awkward 2.x"
)


def test():
    data = np.array([0, 2, 2, 3], dtype="i8")
    offsets = ak.layout.Index64(data)
    offsets = v1_to_v2_index(offsets)

    assert np.asarray(offsets).tolist() == [0, 2, 2, 3]
    assert offsets[0] == 0
    assert offsets[1] == 2
    assert offsets[2] == 2
    assert offsets[3] == 3
    data[2] = 999
    assert offsets[2] == 999

    data = np.array([0, 2, 2, 3], dtype="i4")
    offsets = ak.layout.Index32(data)
    offsets = v1_to_v2_index(offsets)

    assert np.asarray(offsets).tolist() == [0, 2, 2, 3]
    assert offsets[0] == 0
    assert offsets[1] == 2
    assert offsets[2] == 2
    assert offsets[3] == 3
    data[2] = 999
    assert offsets[2] == 999

    content = ak.layout.NumpyArray(np.arange(12).reshape(3, 4))
    data = np.array([0, 2, 2, 3], dtype="i4")
    offsets = ak.layout.Index32(data)
    array = ak.layout.ListOffsetArray32(offsets, content)

    array = v1_to_v2(array)
    content = v1_to_v2(content)

    assert np.asarray(content).tolist() == [[0, 1, 2, 3], [4, 5, 6, 7], [8, 9, 10, 11]]
    assert np.asarray(content[0]).tolist() == [0, 1, 2, 3]
    assert content.typetracer[0].form == content[0].form
    assert np.asarray(content[1]).tolist() == [4, 5, 6, 7]
    assert content.typetracer[1].form == content[1].form
    assert np.asarray(content[2]).tolist() == [8, 9, 10, 11]
    assert content.typetracer[2].form == content[2].form
    assert [content[i][j] for i in range(3) for j in range(4)] == [
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
    ]

    assert np.asarray(array[0]).tolist() == [[0, 1, 2, 3], [4, 5, 6, 7]]
    assert array.typetracer[0].form == array[0].form
    assert np.asarray(array[1]).tolist() == []
    assert array.typetracer[1].form == array[1].form
    assert np.asarray(array[2]).tolist() == [[8, 9, 10, 11]]
    assert array.typetracer[2].form == array[2].form
    assert np.asarray(array[1:3][0]).tolist() == []
    assert array.typetracer[1:3][0].form == array[1:3][0].form
    assert np.asarray(array[1:3][1]).tolist() == [[8, 9, 10, 11]]
    assert array.typetracer[1:3][1].form == array[1:3][1].form
    assert np.asarray(array[2:3][0]).tolist() == [[8, 9, 10, 11]]
    assert array.typetracer[2:3][0].form == array[2:3][0].form


def test_len():
    offsets = ak.layout.Index32(np.array([0, 2, 2, 3], dtype="i4"))
    content = ak.layout.NumpyArray(np.arange(12).reshape(4, 3))
    array = ak.layout.ListOffsetArray32(offsets, content)
    content = v1_to_v2(content)
    array = v1_to_v2(array)

    assert len(content) == 4
    assert len(array) == 3


def test_members():
    offsets = ak.layout.Index32(np.array([0, 2, 2, 3], dtype="i4"))
    content = ak.layout.NumpyArray(np.arange(12).reshape(3, 4))
    array = ak.layout.ListOffsetArray32(offsets, content)
    new = ak.layout.ListOffsetArray32(offsets, array)

    array = v1_to_v2(array)
    new = v1_to_v2(new)

    assert np.asarray(array.offsets).tolist() == [0, 2, 2, 3]
    assert np.asarray(array.content).tolist() == [
        [0, 1, 2, 3],
        [4, 5, 6, 7],
        [8, 9, 10, 11],
    ]

    assert np.asarray(new.offsets).tolist() == [0, 2, 2, 3]
    assert np.asarray(new.content.offsets).tolist() == [0, 2, 2, 3]
    assert np.asarray(new.content.content).tolist() == [
        [0, 1, 2, 3],
        [4, 5, 6, 7],
        [8, 9, 10, 11],
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    array = ak.Array([{"x": 1}, {"x": 2}, None, {"x": 3}])
    assert ak.fill_none(array, array[0], axis=0).tolist() == [
        {"x": 1},
        {"x": 2},
        {"x": 1},
        {"x": 3},
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    array = ak.Array([{"x": 1}])
    array["y"] = ak.virtual(lambda: [2], cache={}, length=1)
    ak.to_list(array)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import os

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401

pytest.importorskip("pyarrow")


def test_list_to_arrow():
    ak_array = ak.Array([[1.1, 2.2, 3.3], [], [4.4, 5.5]])
    pa_array = ak.to_arrow(ak_array)
    assert str(pa_array.type) == "large_list<item: double not null>"
    assert pa_array.to_pylist() == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]

    ak_array = ak.Array(
        ak.layout.ListOffsetArray64(
            ak_array.layout.offsets, ak.layout.UnmaskedArray(ak_array.layout.content)
        )
    )
    pa_array = ak.to_arrow(ak_array)
    assert str(pa_array.type) == "large_list<item: double>"
    assert pa_array.to_pylist() == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]

    ak_array = ak.Array([[1.1, 2.2, None], [], [4.4, 5.5]])
    pa_array = ak.to_arrow(ak_array)
    assert str(pa_array.type) == "large_list<item: double>"
    assert pa_array.to_pylist() == [[1.1, 2.2, None], [], [4.4, 5.5]]


def test_record_to_arrow():
    x_content = ak.Array([1.1, 2.2, 3.3, 4.4, 5.5]).layout
    z_content = ak.Array([1, 2, 3, None, 5]).layout

    ak_array = ak.Array(
        ak.layout.RecordArray(
            [
                x_content,
                ak.layout.UnmaskedArray(x_content),
                z_content,
            ],
            ["x", "y", "z"],
        )
    )
    pa_array = ak.to_arrow(ak_array)
    assert str(pa_array.type) == "struct<x: double not null, y: double, z: int64>"
    assert pa_array.to_pylist() == [
        {"x": 1.1, "y": 1.1, "z": 1},
        {"x": 2.2, "y": 2.2, "z": 2},
        {"x": 3.3, "y": 3.3, "z": 3},
        {"x": 4.4, "y": 4.4, "z": None},
        {"x": 5.5, "y": 5.5, "z": 5},
    ]


def test_union_to_arrow():
    ak_array = ak.Array([1.1, 2.2, None, [1, 2, 3], "hello"])
    pa_array = ak.to_arrow(ak_array)
    assert (
        str(pa_array.type)
        == "dense_union<0: double=0, 1: large_list<item: int64 not null>=1, 2: string=2>"
    )
    assert pa_array.to_pylist() == [1.1, 2.2, None, [1, 2, 3], "hello"]

    ak_array = ak.Array(
        ak.layout.UnmaskedArray(ak.Array([1.1, 2.2, [1, 2, 3], "hello"]).layout)
    )
    pa_array = ak.to_arrow(ak_array)
    assert (
        str(pa_array.type)
        == "dense_union<0: double=0, 1: large_list<item: int64 not null>=1, 2: string=2>"
    )
    assert pa_array.to_pylist() == [1.1, 2.2, [1, 2, 3], "hello"]

    ak_array = ak.Array([1.1, 2.2, [1, 2, 3], "hello"])
    pa_array = ak.to_arrow(ak_array)
    assert (
        str(pa_array.type)
        == "dense_union<0: double not null=0, 1: large_list<item: int64 not null> not null=1, 2: string not null=2>"
    )
    assert pa_array.to_pylist() == [1.1, 2.2, [1, 2, 3], "hello"]


def test_list_from_arrow():
    original = ak.Array([[1.1, 2.2, 3.3], [], [4.4, 5.5]])
    pa_array = ak.to_arrow(original)
    reconstituted = ak.from_arrow(pa_array)
    assert str(reconstituted.type) == "3 * var * float64"
    assert reconstituted.tolist() == [[1.1, 2.2, 3.3], [], [4.4, 5.5]]

    original = ak.Array([[1.1, 2.2, None], [], [4.4, 5.5]])
    pa_array = ak.to_arrow(original)
    reconstituted = ak.from_arrow(pa_array)
    assert str(reconstituted.type) == "3 * var * ?float64"
    assert reconstituted.tolist() == [[1.1, 2.2, None], [], [4.4, 5.5]]

    original = ak.Array([[1.1, 2.2, 3.3], [], None, [4.4, 5.5]])
    pa_array = ak.to_arrow(original)
    reconstituted = ak.from_arrow(pa_array)
    assert str(reconstituted.type) == "4 * option[var * float64]"
    assert reconstituted.tolist() == [[1.1, 2.2, 3.3], [], None, [4.4, 5.5]]

    original = ak.Array([[1.1, 2.2, None], [], None, [4.4, 5.5]])
    pa_array = ak.to_arrow(original)
    reconstituted = ak.from_arrow(pa_array)
    assert str(reconstituted.type) == "4 * option[var * ?float64]"
    assert reconstituted.tolist() == [[1.1, 2.2, None], [], None, [4.4, 5.5]]


def test_record_from_arrow():
    x_content = ak.Array([1.1, 2.2, 3.3, 4.4, 5.5]).layout
    z_content = ak.Array([1, 2, 3, None, 5]).layout

    original = ak.Array(
        ak.layout.RecordArray(
            [
                x_content,
                ak.layout.UnmaskedArray(x_content),
                z_content,
            ],
            ["x", "y", "z"],
        )
    )
    pa_array = ak.to_arrow(original)
    reconstituted = ak.from_arrow(pa_array)
    assert str(reconstituted.type) == '5 * {"x": float64, "y": ?float64, "z": ?int64}'
    assert reconstituted.tolist() == [
        {"x": 1.1, "y": 1.1, "z": 1},
        {"x": 2.2, "y": 2.2, "z": 2},
        {"x": 3.3, "y": 3.3, "z": 3},
        {"x": 4.4, "y": 4.4, "z": None},
        {"x": 5.5, "y": 5.5, "z": 5},
    ]

    original = ak.Array(
        ak.layout.ByteMaskedArray(
            ak.layout.Index8(np.array([False, True, False, False, False], np.int8)),
            original.layout,
            valid_when=False,
        )
    )
    pa_array = ak.to_arrow(original)
    reconstituted = ak.from_arrow(pa_array)
    assert str(reconstituted.type) == '5 * ?{"x": float64, "y": ?float64, "z": ?int64}'
    assert reconstituted.tolist() == [
        {"x": 1.1, "y": 1.1, "z": 1},
        None,
        {"x": 3.3, "y": 3.3, "z": 3},
        {"x": 4.4, "y": 4.4, "z": None},
        {"x": 5.5, "y": 5.5, "z": 5},
    ]


def test_union_from_arrow():
    original = ak.Array([1.1, 2.2, [1, 2, 3], "hello"])
    pa_array = ak.to_arrow(original)
    reconstituted = ak.from_arrow(pa_array)
    assert str(reconstituted.type) == "4 * union[float64, var * int64, string]"
    assert reconstituted.tolist() == [1.1, 2.2, [1, 2, 3], "hello"]

    original = ak.Array([1.1, 2.2, None, [1, 2, 3], "hello"])
    pa_array = ak.to_arrow(original)
    reconstituted = ak.from_arrow(pa_array)
    assert (
        str(reconstituted.type)
        == "5 * union[?float64, option[var * int64], option[string]]"
    )
    assert reconstituted.tolist() == [1.1, 2.2, None, [1, 2, 3], "hello"]


def test_to_arrow_table():
    assert ak.from_arrow(
        ak.to_arrow_table(
            ak.Array([[{"x": 1.1, "y": [1]}], [], [{"x": 2.2, "y": [1, 2]}]]),
            explode_records=True,
        )
    ).tolist() == [
        {"x": [1.1], "y": [[1]]},
        {"x": [], "y": []},
        {"x": [2.2], "y": [[1, 2]]},
    ]
    assert ak.from_arrow(
        ak.to_arrow_table(ak.Array([{"x": 1.1, "y": [1]}, {"x": 2.2, "y": [1, 2]}]))
    ).tolist() == [{"x": 1.1, "y": [1]}, {"x": 2.2, "y": [1, 2]}]


pytest.importorskip("pyarrow.parquet")


def test_to_parquet(tmp_path):
    original = ak.Array(
        [
            [{"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}, {"x": 3, "y": 3.3}],
            [],
            [{"x": 4, "y": 4.4}, {"x": 5, "y": 5.5}],
            [],
            [],
            [
                {"x": 6, "y": 6.6},
                {"x": 7, "y": 7.7},
                {"x": 8, "y": 8.8},
                {"x": 9, "y": 9.9},
            ],
        ]
    )

    ak.to_parquet(original, os.path.join(tmp_path, "data.parquet"))
    reconstituted = ak.from_parquet(os.path.join(tmp_path, "data.parquet"))
    assert reconstituted.tolist() == [
        [{"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}, {"x": 3, "y": 3.3}],
        [],
        [{"x": 4, "y": 4.4}, {"x": 5, "y": 5.5}],
        [],
        [],
        [
            {"x": 6, "y": 6.6},
            {"x": 7, "y": 7.7},
            {"x": 8, "y": 8.8},
            {"x": 9, "y": 9.9},
        ],
    ]
    assert str(reconstituted.type) == '6 * var * {"x": int64, "y": float64}'


def test_to_parquet_2(tmp_path):
    array = ak.Array(
        [
            [{"x": 0.0, "y": []}, {"x": 1.1, "y": [1]}, {"x": 2.2, "y": None}],
            [],
            [{"x": 3.3, "y": [1, 2, 3]}, None, {"x": 4.4, "y": [1, 2, 3, 4]}],
        ]
    )
    assert str(array.type) == '3 * var * ?{"x": float64, "y": option[var * int64]}'
    ak.to_parquet(array, os.path.join(tmp_path, "complicated-example.parquet"))
    array2 = ak.from_parquet(os.path.join(tmp_path, "complicated-example.parquet"))
    assert str(array2.type) == str(array.type)
    assert array2.tolist() == array.tolist()


def test_to_table_2():
    array = ak.Array(
        [
            [{"x": 0.0, "y": []}, {"x": 1.1, "y": [1]}, {"x": 2.2, "y": None}],
            [],
            [{"x": 3.3, "y": [1, 2, 3]}, None, {"x": 4.4, "y": [1, 2, 3, 4]}],
        ]
    )
    assert str(array.type) == '3 * var * ?{"x": float64, "y": option[var * int64]}'
    pa_table = ak.to_arrow_table(array)
    array2 = ak.from_arrow(pa_table)
    assert str(array2.type) == str(array.type)
    assert array2.tolist() == array.tolist()


def test_from_buffers():
    array = ak.Array(
        [
            [{"x": 0.0, "y": []}, {"x": 1.1, "y": [1]}, {"x": 2.2, "y": None}],
            [],
            [{"x": 3.3, "y": [1, 2, 3]}, None, {"x": 4.4, "y": [1, 2, 3, 4]}],
        ]
    )
    assert str(array.type) == '3 * var * ?{"x": float64, "y": option[var * int64]}'
    pa_table = ak.to_arrow_table(array)
    awkward_array = ak.from_arrow(pa_table)
    form, length, container = ak.to_buffers(awkward_array)
    reconstituted = ak.from_buffers(form, length, container, lazy=True)
    assert reconstituted[2].tolist() == [
        {"x": 3.3, "y": [1, 2, 3]},
        None,
        {"x": 4.4, "y": [1, 2, 3, 4]},
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import numbers

import numpy

import awkward.nplike
import awkward._util
import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


class NoError(object):
    str = None
    filename = None
    pass_through = False
    attempt = ak._util.kSliceNone
    id = ak._util.kSliceNone


class NoKernel(object):
    def __call__(self, *args):
        return NoError()


class Interval(object):
    @classmethod
    def unknown(cls):
        return cls(0, None)

    @classmethod
    def exact(cls, value):
        return cls(value, value)

    def __init__(self, min, max):
        assert max is None or min <= max
        self._min = int(min)
        self._max = None if max is None else int(max)

    def __str__(self):
        if self._max is None:
            return "{0}...".format(self._min)
        else:
            return "{0}...{1}".format(self._min, self._max)

    def __repr__(self):
        return "Interval(min={0}, max={1})".format(self._min, self._max)

    @property
    def min(self):
        return self._min

    @property
    def max(self):
        return self._max

    def __index__(self):
        return self._min

    def __eq__(self, other):
        if isinstance(other, Interval):
            return self._min == other._min and self._max == other._max

        elif isinstance(other, numbers.Integral):
            if self._max is None:
                return self._min <= other
            else:
                return self._min <= other <= self._max

    def __add__(self, other):
        if isinstance(other, Interval):
            if self._max is None or other._max is None:
                return Interval(self._min + other._min, None)
            else:
                return Interval(self._min + other._min, self._max + other._max)

        elif isinstance(other, numbers.Integral):
            if self._max is None:
                return Interval(self._min + other, None)
            else:
                return Interval(self._min + other, self._max + other)

        else:
            return NotImplemented

    def __radd__(self, other):
        return self.__add__(other)

    def __iadd__(self, other):
        if isinstance(other, Interval):
            self._min += other._min
            if self._max is not None and other._max is not None:
                self._max += other._max
            else:
                self._max = None

        elif isinstance(other, numbers.Integral):
            self._min += other
            if self._max is not None:
                self._max += other

        else:
            return NotImplemented

        return self

    def __mul__(self, other):
        if isinstance(other, Interval):
            if self._max is None or other._max is None:
                return Interval(self._min * other._min, None)
            else:
                return Interval(self._min * other._min, self._max * other._max)

        elif isinstance(other, numbers.Integral):
            if self._max is None:
                return Interval(self._min * other, None)
            else:
                return Interval(self._min * other, self._max * other)

        else:
            return NotImplemented

    def __rmul__(self, other):
        return self.__mul__(other)

    def __imul__(self, other):
        if isinstance(other, Interval):
            self._min *= other._min
            if self._max is not None and other._max is not None:
                self._max *= other._max
            else:
                self._max = None

        elif isinstance(other, numbers.Integral):
            self._min *= other
            if self._max is not None:
                self._max *= other

        else:
            return NotImplemented

        return self

    def __lt__(self, other):
        if isinstance(other, Interval):
            return NotImplemented

        elif isinstance(other, numbers.Integral):
            return self._min < other

        else:
            return NotImplemented

    def __le__(self, other):
        if isinstance(other, Interval):
            return NotImplemented

        elif isinstance(other, numbers.Integral):
            return self._min <= other

        else:
            return NotImplemented

    def __gt__(self, other):
        if isinstance(other, Interval):
            return NotImplemented

        elif isinstance(other, numbers.Integral):
            return self._min > other

        else:
            return NotImplemented

    def __ge__(self, other):
        if isinstance(other, Interval):
            return NotImplemented

        elif isinstance(other, numbers.Integral):
            return self._min >= other

        else:
            return NotImplemented


def _length_after_slice(slice, original_length):
    start, stop, step = slice.indices(original_length)
    assert step != 0

    if (step > 0 and stop - start > 0) or (step < 0 and stop - start < 0):
        d, m = divmod(abs(start - stop), abs(step))
        return d + (1 if m != 0 else 0)
    else:
        return 0


class TypeTracerArray(object):
    @classmethod
    def from_array(cls, array, dtype=None, fill_zero=0, fill_other=0):
        if isinstance(array, ak._v2.index.Index):
            array = array.data

        if dtype is None:
            dtype = array.dtype

        shape = list(array.shape)
        if len(shape) != 0 and not isinstance(shape[0], Interval):
            shape[0] = Interval.exact(shape[0])

        return cls(dtype, shape=shape, fill_zero=fill_zero, fill_other=fill_other)

    def __init__(self, dtype, shape=None, fill_zero=0, fill_other=0):
        self._dtype = np.dtype(dtype)
        self.shape = shape
        self._fill_zero = fill_zero
        self._fill_other = fill_other

    def __repr__(self):
        dtype = repr(self._dtype)

        shape = ""
        if self._shape != (Interval.unknown(),):
            shape = ", " + repr(self._shape)

        fills = ""
        if self._fill_zero != 0 or self._fill_other != 0:
            fills = ", fill_zero={0}, fill_other={1}".format(
                self._fill_zero, self._fill_other
            )
        return "TypeTracerArray({0}{1}{2})".format(dtype, shape, fills)

    @property
    def dtype(self):
        return self._dtype

    @property
    def shape(self):
        return self._shape

    @shape.setter
    def shape(self, value):
        if value is None:
            value = (Interval.unknown(),)
        elif isinstance(value, Interval):
            value = (value,)
        elif isinstance(value, numbers.Integral):
            value = (Interval.exact(value),)
        else:
            if len(value) == 0:
                value = ()
            elif isinstance(value[0], Interval):
                if not isinstance(value, tuple):
                    value = tuple(value)
            elif isinstance(value[0], numbers.Integral):
                value = (Interval.exact(value[0]),) + tuple(value[1:])
        self._shape = value

    @property
    def strides(self):
        out = (self._dtype.itemsize,)
        for x in self._shape[:0:-1]:
            out = (x * out[0],) + out
        return out

    @property
    def fill_zero(self):
        return self._fill_zero

    @fill_zero.setter
    def fill_zero(self, value):
        self._fill_zero = value

    @property
    def fill_other(self):
        return self._fill_other

    @fill_other.setter
    def fill_other(self, value):
        self._fill_other = value

    @property
    def nplike(self):
        return TypeTracer.instance()

    @property
    def ndim(self):
        return len(self._shape)

    def __iter__(self):
        raise AssertionError(
            "bug in Awkward Array: attempt to convert TypeTracerArray into a concrete array"
        )

    def __array__(self, *args, **kwargs):
        return self.__iter__()

    def itemsize(self):
        return self._dtype.itemsize

    class _CTypes(object):
        data = 0

    @property
    def ctypes(self):
        return self._CTypes

    def __len__(self):
        return self._shape[0]

    def __setitem__(self, where, what):
        pass

    def __getitem__(self, where):
        if isinstance(where, tuple):
            try:
                i = where.index(Ellipsis)
            except ValueError:
                pass
            else:
                before, after = where[:i], where[i + 1 :]
                missing = max(0, len(self._shape) - (len(before) + len(after)))
                where = before + (slice(None, None, None),) * missing + after

        if isinstance(where, numbers.Integral):
            if len(self._shape) == 1:
                if where == 0:
                    return self._dtype.type(self._fill_zero)
                else:
                    return self._dtype.type(self._fill_other)
            else:
                out = numpy.full(self._shape[1:], self._fill_other, dtype=self._dtype)
                out[0] = self._fill_zero
                return out

        elif isinstance(where, slice):
            length1 = _length_after_slice(where, self._shape[0].min)

            if self._shape[0].max is not None:
                length2 = _length_after_slice(where, self._shape[0].max)
            else:
                if where.start is not None and where.stop is not None:
                    length2 = _length_after_slice(
                        where, max(abs(where.start), abs(where.stop))
                    )
                elif where.stop is not None:
                    length2 = _length_after_slice(abs(where.stop))
                else:
                    length2 = None

            shape = (Interval(length1, length2),) + self._shape[1:]

            return TypeTracerArray(
                self._dtype, shape, self._fill_zero, self._fill_other
            )

        elif (
            hasattr(where, "dtype")
            and hasattr(where, "shape")
            and issubclass(where.dtype.type, np.integer)
        ):
            assert len(self._shape) != 0
            shape = where.shape + self._shape[1:]
            return TypeTracerArray(
                self._dtype, shape, self._fill_zero, self._fill_other
            )

        elif isinstance(where, tuple) and any(
            hasattr(x, "dtype") and hasattr(x, "shape") for x in where
        ):
            for num_basic, wh in enumerate(where):  # noqa: B007
                if not isinstance(wh, slice):
                    break

            if num_basic != 0:
                tmp = self.__getitem__(where[:num_basic])
                basic_shape = tmp._shape[:num_basic]
            else:
                basic_shape = ()

            shapes = []
            for j in range(num_basic, len(where)):
                wh = where[j]
                if isinstance(wh, numbers.Integral):
                    shapes.append(numpy.array(0))
                elif hasattr(wh, "dtype") and hasattr(wh, "shape"):
                    sh = [
                        x.min if isinstance(x, Interval) else int(x) for x in wh.shape
                    ]
                    shapes.append(
                        numpy.lib.stride_tricks.as_strided(
                            numpy.array(0), shape=sh, strides=[0] * len(sh)
                        )
                    )
                else:
                    raise NotImplementedError(repr(wh))

            slicer_shape = numpy.broadcast_arrays(*shapes)[0].shape

            shape = basic_shape + slicer_shape + self._shape[num_basic + len(shapes) :]
            if len(shape) == 0 or isinstance(shape[0], Interval):
                fixed_shape = shape
            else:
                fixed_shape = (Interval.exact(shape[0]),) + shape[1:]

            return TypeTracerArray(
                self._dtype, fixed_shape, self._fill_zero, self._fill_other
            )

        elif (
            isinstance(where, tuple)
            and len(where) > 0
            and isinstance(where[0], (numbers.Integral, slice))
        ):
            head, tail = where[0], where[1:]
            next = self.__getitem__(head)

            inner_shape = next.shape[1:]
            after_shape = []
            for i, wh in enumerate(tail):
                if isinstance(wh, int):
                    pass
                elif isinstance(wh, slice):
                    after_shape.append(_length_after_slice(wh, inner_shape[i]))
                else:
                    raise NotImplementedError(repr(wh))

            shape = (next._shape[0],) + tuple(after_shape)
            return TypeTracerArray(
                self._dtype, shape, self._fill_zero, self._fill_other
            )

        else:
            raise NotImplementedError(repr(where))

    def __lt__(self, other):
        if isinstance(other, numbers.Real):
            return TypeTracerArray(np.bool_, self._shape, False, False)
        else:
            return NotImplemented

    def __le__(self, other):
        if isinstance(other, numbers.Real):
            return TypeTracerArray(np.bool_, self._shape, False, False)
        else:
            return NotImplemented

    def __gt__(self, other):
        if isinstance(other, numbers.Real):
            return TypeTracerArray(np.bool_, self._shape, False, False)
        else:
            return NotImplemented

    def __ge__(self, other):
        if isinstance(other, numbers.Real):
            return TypeTracerArray(np.bool_, self._shape, False, False)
        else:
            return NotImplemented

    def reshape(self, *args):
        if len(args) == 1 and isinstance(args[0], tuple):
            args = args[0]

        assert len(args) != 0
        assert all(isinstance(x, numbers.Integral) for x in args)
        assert all(x >= 0 for x in args[1:])

        if args[0] < 0:
            minitems, maxitems = self._shape[0].min, self._shape[0].max
            for x in self._shape[1:]:
                minitems *= x
                if maxitems is not None:
                    maxitems *= x

            divisor = 1
            for x in args[1:]:
                divisor *= x

            minlength = minitems // divisor
            if maxitems is None:
                maxlength = None
            else:
                maxlength = maxitems // divisor

            shape = (Interval(minlength, maxlength),) + args[1:]

        else:
            shape = (Interval.exact(args[0]),) + args[1:]

        return TypeTracerArray(self._dtype, shape, self._fill_zero, self._fill_other)

    def copy(self):
        return self


unset = object()


class TypeTracer(ak.nplike.NumpyLike):
    known_shape = False

    def to_rectilinear(self, array, *args, **kwargs):
        raise NotImplementedError

    def __getitem__(self, name_and_types):
        return NoKernel()

    @property
    def ma(self):
        raise NotImplementedError

    @property
    def char(self):
        raise NotImplementedError

    @property
    def ndarray(self):
        return TypeTracerArray

    ############################ array creation

    def array(self, data, dtype=unset, **kwargs):
        # data[, dtype=[, copy=]]
        if dtype is unset:
            dtype = data.dtype
        return TypeTracerArray.from_array(data, dtype=dtype)

    def asarray(self, array, dtype=unset, **kwargs):
        # array[, dtype=][, order=]
        if dtype is unset:
            dtype = array.dtype
        return TypeTracerArray.from_array(array, dtype=dtype)

    def ascontiguousarray(self, array, dtype=unset, **kwargs):
        # array[, dtype=]
        if dtype is unset:
            dtype = array.dtype
        return TypeTracerArray.from_array(array, dtype=dtype)

    def isscalar(self, *args, **kwargs):
        raise NotImplementedError

    def frombuffer(self, *args, **kwargs):
        # array[, dtype=]
        raise NotImplementedError

    def zeros(self, shape, dtype=np.float64, **kwargs):
        # shape/len[, dtype=]
        return TypeTracerArray(dtype, shape)

    def ones(self, shape, dtype=np.float64, **kwargs):
        # shape/len[, dtype=]
        return TypeTracerArray(dtype, shape)

    def empty(self, shape, dtype=np.float64, **kwargs):
        # shape/len[, dtype=]
        return TypeTracerArray(dtype, shape)

    def full(self, shape, value, dtype=unset, **kwargs):
        # shape/len, value[, dtype=]
        if dtype is unset:
            dtype = numpy.array(value).dtype
        return TypeTracerArray(dtype, shape, fill_zero=value, fill_other=value)

    def zeros_like(self, *args, **kwargs):
        # array
        raise NotImplementedError

    def ones_like(self, *args, **kwargs):
        # array
        raise NotImplementedError

    def full_like(self, *args, **kwargs):
        # array, fill_value
        raise NotImplementedError

    def arange(self, *args, **kwargs):
        # stop[, dtype=]
        # start, stop[, dtype=]
        # start, stop, step[, dtype=]
        raise NotImplementedError

    def meshgrid(self, *args, **kwargs):
        # *arrays, indexing="ij"
        raise NotImplementedError

    ############################ testing

    def shape(self, *args, **kwargs):
        # array
        raise NotImplementedError

    def array_equal(self, *args, **kwargs):
        # array1, array2
        raise NotImplementedError

    def size(self, *args, **kwargs):
        # array
        raise NotImplementedError

    def searchsorted(self, *args, **kwargs):
        # haystack, needle, side="right"
        raise NotImplementedError

    def argsort(self, *args, **kwargs):
        # array
        raise NotImplementedError

    ############################ manipulation

    def broadcast_arrays(self, *args, **kwargs):
        # array1[, array2[, ...]]
        raise NotImplementedError

    def add(self, *args, **kwargs):
        # array1, array2[, out=]
        raise NotImplementedError

    def cumsum(self, *args, **kwargs):
        # arrays[, out=]
        raise NotImplementedError

    def cumprod(self, *args, **kwargs):
        # arrays[, out=]
        raise NotImplementedError

    def nonzero(self, *args, **kwargs):
        # array
        raise NotImplementedError

    def unique(self, *args, **kwargs):
        # array
        raise NotImplementedError

    def concatenate(self, *args, **kwargs):
        # arrays
        raise NotImplementedError

    def repeat(self, *args, **kwargs):
        # array, int
        # array1, array2
        raise NotImplementedError

    def stack(self, *args, **kwargs):
        # arrays
        raise NotImplementedError

    def vstack(self, *args, **kwargs):
        # arrays
        raise NotImplementedError

    def packbits(self, *args, **kwargs):
        # array
        raise NotImplementedError

    def unpackbits(self, *args, **kwargs):
        # array
        raise NotImplementedError

    def atleast_1d(self, *args, **kwargs):
        # *arrays
        raise NotImplementedError

    def broadcast_to(self, *args, **kwargs):
        # array, shape
        raise NotImplementedError

    def append(self, *args, **kwargs):
        # array, element
        raise NotImplementedError

    def where(self, *args, **kwargs):
        # array, element
        raise NotImplementedError

    ############################ ufuncs

    def sqrt(self, *args, **kwargs):
        # array
        raise NotImplementedError

    def exp(self, *args, **kwargs):
        # array
        raise NotImplementedError

    def true_divide(self, *args, **kwargs):
        # array1, array2
        raise NotImplementedError

    def bitwise_or(self, *args, **kwargs):
        # array1, array2[, out=output]
        raise NotImplementedError

    def logical_and(self, *args, **kwargs):
        # array1, array2
        raise NotImplementedError

    def equal(self, *args, **kwargs):
        # array1, array2
        raise NotImplementedError

    def ceil(self, *args, **kwargs):
        # array
        raise NotImplementedError

    ############################ almost-ufuncs

    def nan_to_num(self, *args, **kwargs):
        # array, copy=True, nan=0.0, posinf=None, neginf=None
        raise NotImplementedError

    def isclose(self, *args, **kwargs):
        # a, b, rtol=1e-05, atol=1e-08, equal_nan=False
        raise NotImplementedError

    ############################ reducers

    def all(self, array, prefer):
        # array
        return prefer

    def any(self, array, prefer):
        # array
        return prefer

    def count_nonzero(self, *args, **kwargs):
        # array
        raise NotImplementedError

    def sum(self, *args, **kwargs):
        # array
        raise NotImplementedError

    def prod(self, *args, **kwargs):
        # array
        raise NotImplementedError

    def min(self, *args, **kwargs):
        # array
        raise NotImplementedError

    def max(self, *args, **kwargs):
        # array
        raise NotImplementedError

    def argmin(self, *args, **kwargs):
        # array[, axis=]
        raise NotImplementedError

    def argmax(self, *args, **kwargs):
        # array[, axis=]
        raise NotImplementedError

    def array_str(
        self, array, max_line_width=unset, precision=unset, suppress_small=unset
    ):
        # array, max_line_width, precision=None, suppress_small=None
        return "[?? ... ??]"

    def datetime_as_string(self, *args, **kwargs):
        raise NotImplementedError



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import json

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


class Type(object):
    @property
    def parameters(self):
        return self._parameters

    def parameter(self, key):
        if self._parameters is None:
            return None
        else:
            return self._parameters.get(key)

    @property
    def typestr(self):
        return self._typestr

    _str_parameters_exclude = ("__categorical__",)

    def _str_categorical_begin(self):
        if self.parameter("__categorical__") is not None:
            return "categorical[type="
        else:
            return ""

    def _str_categorical_end(self):
        if self.parameter("__categorical__") is not None:
            return "]"
        else:
            return ""

    def _str_parameters(self):
        out = []
        if self._parameters is not None:
            for k, v in self._parameters.items():
                if k not in self._str_parameters_exclude:
                    out.append(json.dumps(k) + ": " + json.dumps(v))

        if len(out) == 0:
            return None
        else:
            return "parameters={" + ", ".join(out) + "}"

    def _repr_args(self):
        out = []

        if self._parameters is not None and len(self._parameters) > 0:
            out.append("parameters=" + repr(self._parameters))

        if self._typestr is not None:
            out.append("typestr=" + repr(self._typestr))

        return out



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test_slice():
    assert ak._ext._slice_tostring(3) == "[3]"
    assert ak._ext._slice_tostring(slice(None)) == "[:]"
    assert ak._ext._slice_tostring(slice(10)) == "[:10]"
    assert ak._ext._slice_tostring(slice(1, 2)) == "[1:2]"
    assert ak._ext._slice_tostring(slice(1, None)) == "[1:]"
    assert ak._ext._slice_tostring(slice(None, None, 2)) == "[::2]"
    assert ak._ext._slice_tostring(slice(1, 2, 3)) == "[1:2:3]"
    assert ak._ext._slice_tostring(Ellipsis) == "[...]"
    assert ak._ext._slice_tostring(np.newaxis) == "[newaxis]"
    assert ak._ext._slice_tostring(None) == "[newaxis]"
    assert ak._ext._slice_tostring([1, 2, 3]) == "[array([1, 2, 3])]"
    assert (
        ak._ext._slice_tostring(np.array([[1, 2], [3, 4], [5, 6]]))
        == "[array([[1, 2], [3, 4], [5, 6]])]"
    )
    assert (
        ak._ext._slice_tostring(np.array([1, 2, 3, 4, 5, 6])[::-2])
        == "[array([6, 4, 2])]"
    )
    a = np.arange(3 * 5).reshape(3, 5)[1::, ::-2]
    assert ak._ext._slice_tostring(a) == "[array(" + str(a.tolist()) + ")]"
    a = np.arange(3 * 5).reshape(3, 5)[::-1, ::2]
    assert ak._ext._slice_tostring(a) == "[array(" + str(a.tolist()) + ")]"
    assert (
        ak._ext._slice_tostring([True, True, False, False, True])
        == "[array([0, 1, 4])]"
    )
    assert (
        ak._ext._slice_tostring([[True, True], [False, False], [True, False]])
        == "[array([0, 0, 2]), array([0, 1, 0])]"
    )
    assert ak._ext._slice_tostring(()) == "[]"
    assert ak._ext._slice_tostring((3,)) == "[3]"
    assert ak._ext._slice_tostring((3, slice(1, 2, 3))) == "[3, 1:2:3]"
    assert ak._ext._slice_tostring((slice(None), [1, 2, 3])) == "[:, array([1, 2, 3])]"
    assert ak._ext._slice_tostring(([1, 2, 3], slice(None))) == "[array([1, 2, 3]), :]"
    assert (
        ak._ext._slice_tostring((slice(None), [True, True, False, False, True]))
        == "[:, array([0, 1, 4])]"
    )
    assert (
        ak._ext._slice_tostring(
            (slice(None), [[True, True], [False, False], [True, False]])
        )
        == "[:, array([0, 0, 2]), array([0, 1, 0])]"
    )
    assert (
        ak._ext._slice_tostring(
            ([[True, True], [False, False], [True, False]], slice(None))
        )
        == "[array([0, 0, 2]), array([0, 1, 0]), :]"
    )

    with pytest.raises(ValueError):
        ak._ext._slice_tostring(np.array([1.1, 2.2, 3.3]))
    assert (
        ak._ext._slice_tostring(np.array(["one", "two", "three"]))
        == '[["one", "two", "three"]]'
    )
    with pytest.raises(ValueError):
        ak._ext._slice_tostring(np.array([1, 2, 3, None, 4, 5]))

    assert (
        ak._ext._slice_tostring((123, [[1, 1], [2, 2], [3, 3]]))
        == "[array([[123, 123], [123, 123], [123, 123]]), array([[1, 1], [2, 2], [3, 3]])]"
    )
    assert (
        ak._ext._slice_tostring(([[1, 1], [2, 2], [3, 3]], 123))
        == "[array([[1, 1], [2, 2], [3, 3]]), array([[123, 123], [123, 123], [123, 123]])]"
    )
    assert (
        ak._ext._slice_tostring(([[100, 200, 300, 400]], [[1], [2], [3]]))
        == "[array([[100, 200, 300, 400], [100, 200, 300, 400], [100, 200, 300, 400]]), array([[1, 1, 1, 1], [2, 2, 2, 2], [3, 3, 3, 3]])]"
    )
    assert (
        ak._ext._slice_tostring(([[1], [2], [3]], [[100, 200, 300, 400]]))
        == "[array([[1, 1, 1, 1], [2, 2, 2, 2], [3, 3, 3, 3]]), array([[100, 200, 300, 400], [100, 200, 300, 400], [100, 200, 300, 400]])]"
    )

    with pytest.raises(ValueError):
        ak._ext._slice_tostring((3, slice(None), [[1], [2], [3]]))
    with pytest.raises(ValueError):
        ak._ext._slice_tostring(([[1, 2, 3, 4]], slice(None), [[1], [2], [3]]))
    with pytest.raises(ValueError):
        ak._ext._slice_tostring(
            (slice(None), 3, slice(None), [[1], [2], [3]], slice(None))
        )
    with pytest.raises(ValueError):
        ak._ext._slice_tostring(
            (slice(None), [[1, 2, 3, 4]], slice(None), [[1], [2], [3]], slice(None))
        )
    assert (
        ak._ext._slice_tostring((slice(None), 3, [[1], [2], [3]], slice(None)))
        == "[:, array([[3], [3], [3]]), array([[1], [2], [3]]), :]"
    )
    assert (
        ak._ext._slice_tostring(
            (slice(None), [[1, 2, 3, 4]], [[1], [2], [3]], slice(None))
        )
        == "[:, array([[1, 2, 3, 4], [1, 2, 3, 4], [1, 2, 3, 4]]), array([[1, 1, 1, 1], [2, 2, 2, 2], [3, 3, 3, 3]]), :]"
    )


def test_numpyarray_getitem_bystrides():
    a = np.arange(10)
    b = ak.layout.NumpyArray(a)
    assert b[3] == a[3]
    assert b[-3] == a[-3]
    assert ak.to_list(b[()]) == ak.to_list(a[()])
    assert ak.to_list(b[slice(None)]) == ak.to_list(a[slice(None)])
    assert ak.to_list(b[slice(3, 7)]) == ak.to_list(a[slice(3, 7)])
    assert ak.to_list(b[slice(3, 100)]) == ak.to_list(a[slice(3, 100)])
    assert ak.to_list(b[slice(-100, 7)]) == ak.to_list(a[slice(-100, 7)])
    assert ak.to_list(b[slice(3, -3)]) == ak.to_list(a[slice(3, -3)])
    assert ak.to_list(b[slice(1, 7, 2)]) == ak.to_list(a[slice(1, 7, 2)])
    assert ak.to_list(b[slice(-8, 7, 2)]) == ak.to_list(a[slice(-8, 7, 2)])
    assert ak.to_list(b[slice(None, 7, 2)]) == ak.to_list(a[slice(None, 7, 2)])
    assert ak.to_list(b[slice(None, -3, 2)]) == ak.to_list(a[slice(None, -3, 2)])
    assert ak.to_list(b[slice(8, None, -1)]) == ak.to_list(a[slice(8, None, -1)])
    assert ak.to_list(b[slice(8, None, -2)]) == ak.to_list(a[slice(8, None, -2)])
    assert ak.to_list(b[slice(-2, None, -2)]) == ak.to_list(a[slice(-2, None, -2)])

    a = np.arange(7 * 5).reshape(7, 5)
    b = ak.layout.NumpyArray(a)

    assert ak.to_list(b[()]) == ak.to_list(a[()])
    assert ak.to_list(b[3]) == ak.to_list(a[3])
    assert ak.to_list(b[(3, 2)]) == ak.to_list(a[3, 2])
    assert ak.to_list(b[slice(1, 4)]) == ak.to_list(a[slice(1, 4)])
    assert ak.to_list(b[(3, slice(1, 4))]) == ak.to_list(a[3, slice(1, 4)])
    assert ak.to_list(b[(slice(1, 4), 3)]) == ak.to_list(a[slice(1, 4), 3])
    assert ak.to_list(b[(slice(1, 4), slice(2, None))]) == ak.to_list(
        a[slice(1, 4), slice(2, None)]
    )
    assert ak.to_list(b[(slice(None, None, -1), slice(2, None))]) == ak.to_list(
        a[slice(None, None, -1), slice(2, None)]
    )

    assert ak.to_list(b[:, np.newaxis, :]) == ak.to_list(a[:, np.newaxis, :])
    assert ak.to_list(b[np.newaxis, :, np.newaxis, :, np.newaxis]) == ak.to_list(
        a[np.newaxis, :, np.newaxis, :, np.newaxis]
    )

    if not ak._util.py27:
        assert ak.to_list(b[Ellipsis, 3]) == ak.to_list(a[Ellipsis, 3])
        assert ak.to_list(b[Ellipsis, 3, 2]) == ak.to_list(a[Ellipsis, 3, 2])
        assert ak.to_list(b[3, Ellipsis]) == ak.to_list(a[3, Ellipsis])
        assert ak.to_list(b[3, 2, Ellipsis]) == ak.to_list(a[3, 2, Ellipsis])


def test_numpyarray_contiguous():
    a = np.arange(10)[8::-2]
    b = ak.layout.NumpyArray(a)

    assert ak.to_list(b) == ak.to_list(a)
    assert ak.to_list(b.contiguous()) == ak.to_list(a)
    b = b.contiguous()
    assert ak.to_list(b) == ak.to_list(a)

    a = np.arange(7 * 5).reshape(7, 5)[::-1, ::2]
    b = ak.layout.NumpyArray(a)

    assert ak.to_list(b) == ak.to_list(a)
    assert ak.to_list(b.contiguous())
    b = b.contiguous()
    assert ak.to_list(b) == ak.to_list(a)


def test_numpyarray_getitem_next():
    a = np.arange(10)
    b = ak.layout.NumpyArray(a)
    c = np.array([7, 3, 3, 5])
    assert ak.to_list(b[c]) == ak.to_list(a[c])
    c = np.array([False, False, False, True, True, True, False, True, False, True])
    assert ak.to_list(b[c]) == ak.to_list(a[c])
    c = np.array([], dtype=int)
    assert ak.to_list(b[c]) == ak.to_list(a[c])

    a = np.arange(10 * 3).reshape(10, 3)
    b = ak.layout.NumpyArray(a)
    c = np.array([7, 3, 3, 5])
    assert ak.to_list(b[c]) == ak.to_list(a[c])
    c = np.array([False, False, False, True, True, True, False, True, False, True])
    assert ak.to_list(b[c]) == ak.to_list(a[c])
    c = np.array([], dtype=int)
    assert ak.to_list(b[c]) == ak.to_list(a[c])

    a = np.arange(7 * 5).reshape(7, 5)
    b = ak.layout.NumpyArray(a)
    c1 = np.array([], np.int64)
    c2 = np.array([], np.int64)
    assert ak.to_list(b[c1, c2]) == ak.to_list(a[c1, c2])
    assert ak.Array(b[c1, c2]).ndim == a[c1, c2].ndim

    a = np.arange(7 * 5).reshape(7, 5)
    b = ak.layout.NumpyArray(a)
    c1 = np.array([4, 1, 1, 3])
    c2 = np.array([2, 2, 0, 1])
    assert ak.to_list(b[c1, c2]) == ak.to_list(a[c1, c2])

    c1 = np.array([[4, 1], [1, 3], [0, 4]])
    c2 = np.array([[2, 2], [0, 1], [1, 3]])
    assert ak.to_list(b[c1, c2]) == ak.to_list(a[c1, c2])
    c = np.array([False, False, True, True, False, True, True])
    assert ak.to_list(b[c]) == ak.to_list(a[c])
    c = a % 2 == 0  # two dimensional
    assert ak.to_list(b[c]) == ak.to_list(a[c])
    c = np.array([], dtype=int)
    assert ak.to_list(b[c]) == ak.to_list(a[c])
    c1 = np.array([], dtype=int)
    c2 = np.array([], dtype=int)
    assert ak.to_list(b[c1, c2]) == ak.to_list(a[c1, c2])

    a = np.arange(7 * 5).reshape(7, 5)
    b = ak.layout.NumpyArray(a)
    c = np.array([2, 0, 0, 1])
    assert ak.to_list(b[1:4, c]) == ak.to_list(a[1:4, c])
    assert ak.to_list(b[c, 1:4]) == ak.to_list(a[c, 1:4])
    assert ak.to_list(b[1:4, np.newaxis, c]) == ak.to_list(a[1:4, np.newaxis, c])
    assert ak.to_list(b[c, np.newaxis, 1:4]) == ak.to_list(a[c, np.newaxis, 1:4])
    assert ak.to_list(b[1:4, np.newaxis, np.newaxis, c]) == ak.to_list(
        a[1:4, np.newaxis, np.newaxis, c]
    )
    assert ak.to_list(b[c, np.newaxis, np.newaxis, 1:4]) == ak.to_list(
        a[c, np.newaxis, np.newaxis, 1:4]
    )
    if not ak._util.py27:
        assert ak.to_list(b[Ellipsis, c]) == ak.to_list(a[Ellipsis, c])
        assert ak.to_list(b[c, Ellipsis]) == ak.to_list(a[c, Ellipsis])



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import os

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


pytest.importorskip("pyarrow.parquet")


def test(tmp_path):
    filename = os.path.join(tmp_path, "test.parquet")
    ak.to_parquet(ak.repartition(range(8), 2), filename)

    assert ak.from_parquet(filename, row_groups=[1, 3]).tolist() == [2, 3, 6, 7]
    assert ak.from_parquet(filename, row_groups=[1, 3], lazy=True).tolist() == [
        2,
        3,
        6,
        7,
    ]

    assert ak.from_parquet(tmp_path, row_groups=[1, 3]).tolist() == [2, 3, 6, 7]
    assert ak.from_parquet(tmp_path, row_groups=[1, 3], lazy=True).tolist() == [
        2,
        3,
        6,
        7,
    ]

    ak.to_parquet.dataset(tmp_path)

    assert ak.from_parquet(tmp_path, row_groups=[1, 3]).tolist() == [2, 3, 6, 7]
    assert ak.from_parquet(tmp_path, row_groups=[1, 3], lazy=True).tolist() == [
        2,
        3,
        6,
        7,
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    a = ak.layout.NumpyArray(np.empty(0))
    idx = ak.layout.Index64([])
    a = ak.layout.IndexedOptionArray64(idx, a)
    idx = ak.layout.Index64([0])
    a = ak.layout.ListOffsetArray64(idx, a)
    idx = ak.layout.Index64([175990832])
    a = ak.layout.ListOffsetArray64(idx, a)
    assert ak.flatten(a, axis=2).tolist() == []
    assert str(ak.flatten(a, axis=2).type) == "0 * var * ?float64"



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import os

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


pytest.importorskip("pyarrow.parquet")


@pytest.mark.parametrize("one,two,three", [(1, 2, 3), ("one", "two", "three")])
def test_1(one, two, three, tmp_path):
    filename = os.path.join(str(tmp_path), "test1.parquet")
    data = [{"x": one}, {"x": two}, {"x": three}]
    ak.to_parquet(ak.Array(data), filename)
    array = ak.from_parquet(filename, lazy=True, lazy_cache_key="tmp")
    assert set(array.caches[0].keys()) == set()
    array.layout.field("x").array
    assert set(array.caches[0].keys()) == set(["tmp:col:x[0]"])
    assert array.tolist() == data


@pytest.mark.parametrize("one,two,three", [(1, 2, 3), ("one", "two", "three")])
def test_2(one, two, three, tmp_path):
    filename = os.path.join(str(tmp_path), "test2.parquet")
    data = [{"x": {"y": one}}, {"x": {"y": two}}, {"x": {"y": three}}]
    ak.to_parquet(ak.Array(data), filename)
    array = ak.from_parquet(filename, lazy=True, lazy_cache_key="tmp")
    assert set(array.caches[0].keys()) == set()
    array.layout.field("x").array
    assert set(array.caches[0].keys()) == set()
    array.layout.field("x").array.field("y").array
    assert set(array.caches[0].keys()) == set(["tmp:col:x.y[0]"])
    assert array.tolist() == data


@pytest.mark.parametrize("one,two,three", [(1, 2, 3), ("one", "two", "three")])
def test_3(one, two, three, tmp_path):
    filename = os.path.join(str(tmp_path), "test3.parquet")
    data = [
        {"x": {"y": one, "z": 1.1}},
        {"x": {"y": two, "z": 2.2}},
        {"x": {"y": three, "z": 3.3}},
    ]
    ak.to_parquet(ak.Array(data), filename)
    array = ak.from_parquet(filename, lazy=True, lazy_cache_key="tmp")
    assert set(array.caches[0].keys()) == set()
    array.layout.field("x").array
    assert set(array.caches[0].keys()) == set()
    array.layout.field("x").array.field("z").array
    assert set(array.caches[0].keys()) == set(["tmp:col:x.z[0]"])
    array.layout.field("x").array.field("y").array
    assert set(array.caches[0].keys()) == set(["tmp:col:x.z[0]", "tmp:col:x.y[0]"])
    assert array.tolist() == data


@pytest.mark.parametrize("one,two,three", [(1, 2, 3), ("one", "two", "three")])
def test_4(one, two, three, tmp_path):
    filename = os.path.join(str(tmp_path), "test4.parquet")
    data = [{"x": []}, {"x": [one]}, {"x": [one, two, three]}]
    ak.to_parquet(ak.Array(data), filename)
    array = ak.from_parquet(filename, lazy=True, lazy_cache_key="tmp")
    assert set(array.caches[0].keys()) == set()
    array.layout.field("x").array
    assert set(array.caches[0].keys()) == set(["tmp:lst:x[0]"])
    assert array.tolist() == data


@pytest.mark.parametrize("one,two,three", [(1, 2, 3), ("one", "two", "three")])
def test_5(one, two, three, tmp_path):
    filename = os.path.join(str(tmp_path), "test5.parquet")
    data = [{"x": {"y": []}}, {"x": {"y": [one]}}, {"x": {"y": [one, two, three]}}]
    ak.to_parquet(ak.Array(data), filename)
    array = ak.from_parquet(filename, lazy=True, lazy_cache_key="tmp")
    assert set(array.caches[0].keys()) == set()
    array.layout.field("x").array
    assert set(array.caches[0].keys()) == set()
    array.layout.field("x").array.field("y").array
    assert set(array.caches[0].keys()) == set(["tmp:lst:x.y[0]"])
    assert array.tolist() == data


@pytest.mark.parametrize("one,two,three", [(1, 2, 3), ("one", "two", "three")])
def test_6(one, two, three, tmp_path):
    filename = os.path.join(str(tmp_path), "test6.parquet")
    data = [
        {"x": {"y": [], "z": 1.1}},
        {"x": {"y": [one], "z": 2.2}},
        {"x": {"y": [one, two, three], "z": 3.3}},
    ]
    ak.to_parquet(ak.Array(data), filename)
    array = ak.from_parquet(filename, lazy=True, lazy_cache_key="tmp")
    assert set(array.caches[0].keys()) == set()
    array.layout.field("x").array
    assert set(array.caches[0].keys()) == set()
    array.layout.field("x").array.field("z").array
    assert set(array.caches[0].keys()) == set(["tmp:col:x.z[0]"])
    array.layout.field("x").array.field("y").array
    assert set(array.caches[0].keys()) == set(["tmp:col:x.z[0]", "tmp:lst:x.y[0]"])
    assert array.tolist() == data
    array = ak.from_parquet(filename, lazy=True, lazy_cache_key="tmp")
    assert set(array.caches[0].keys()) == set()
    array.layout.field("x").array
    assert set(array.caches[0].keys()) == set()
    array.layout.field("x").array.field("y").array
    assert set(array.caches[0].keys()) == set(["tmp:lst:x.y[0]"])
    array.layout.field("x").array.field("z").array
    assert set(array.caches[0].keys()) == set(["tmp:lst:x.y[0]", "tmp:col:x.z[0]"])
    assert array.tolist() == data


@pytest.mark.parametrize("one,two,three", [(1, 2, 3), ("one", "two", "three")])
def test_7(one, two, three, tmp_path):
    filename = os.path.join(str(tmp_path), "test7.parquet")
    data = [
        {"x": []},
        {"x": [{"y": one}]},
        {"x": [{"y": one}, {"y": two}, {"y": three}]},
    ]
    ak.to_parquet(ak.Array(data), filename)
    array = ak.from_parquet(filename, lazy=True, lazy_cache_key="tmp")
    assert set(array.caches[0].keys()) == set()
    array.layout.field("x").array
    assert set(array.caches[0].keys()) == set(["tmp:off:x.list.item.y:x[0]"])
    assert np.asarray(array.layout.field("x").array.offsets).tolist() == [0, 0, 1, 4]
    assert set(array.caches[0].keys()) == set(["tmp:off:x.list.item.y:x[0]"])
    array.layout.field("x").array.content.field("y").array
    assert set(array.caches[0].keys()) == set(
        ["tmp:off:x.list.item.y:x[0]", "tmp:col:x.list.item.y[0]"]
    )
    assert array.tolist() == data


@pytest.mark.parametrize("one,two,three", [(1, 2, 3), ("one", "two", "three")])
def test_8(one, two, three, tmp_path):
    filename = os.path.join(str(tmp_path), "test8.parquet")
    data = [
        {"x": []},
        {"x": [{"y": one, "z": 1.1}]},
        {"x": [{"y": one, "z": 1.1}, {"y": two, "z": 2.2}, {"y": three, "z": 3.3}]},
    ]
    ak.to_parquet(ak.Array(data), filename)
    array = ak.from_parquet(filename, lazy=True, lazy_cache_key="tmp")
    assert set(array.caches[0].keys()) == set()
    array.layout.field("x").array
    assert set(array.caches[0].keys()) == set(["tmp:off:x.list.item.y:x[0]"])
    assert np.asarray(array.layout.field("x").array.offsets).tolist() == [0, 0, 1, 4]
    assert set(array.caches[0].keys()) == set(["tmp:off:x.list.item.y:x[0]"])
    array.layout.field("x").array.content.field("y").array
    assert set(array.caches[0].keys()) == set(
        ["tmp:off:x.list.item.y:x[0]", "tmp:col:x.list.item.y[0]"]
    )
    array.layout.field("x").array.content.field("z").array
    assert set(array.caches[0].keys()) == set(
        [
            "tmp:off:x.list.item.y:x[0]",
            "tmp:col:x.list.item.y[0]",
            "tmp:col:x.list.item.z[0]",
        ]
    )
    assert array.tolist() == data
    array = ak.from_parquet(filename, lazy=True, lazy_cache_key="tmp")
    assert set(array.caches[0].keys()) == set()
    array.layout.field("x").array
    assert set(array.caches[0].keys()) == set(["tmp:off:x.list.item.y:x[0]"])
    assert np.asarray(array.layout.field("x").array.offsets).tolist() == [0, 0, 1, 4]
    assert set(array.caches[0].keys()) == set(["tmp:off:x.list.item.y:x[0]"])
    array.layout.field("x").array.content.field("z").array
    assert set(array.caches[0].keys()) == set(
        ["tmp:off:x.list.item.y:x[0]", "tmp:col:x.list.item.z[0]"]
    )
    array.layout.field("x").array.content.field("y").array
    assert set(array.caches[0].keys()) == set(
        [
            "tmp:off:x.list.item.y:x[0]",
            "tmp:col:x.list.item.z[0]",
            "tmp:col:x.list.item.y[0]",
        ]
    )
    assert array.tolist() == data


@pytest.mark.parametrize("one,two,three", [(1, 2, 3), ("one", "two", "three")])
def test_9(one, two, three, tmp_path):
    filename = os.path.join(str(tmp_path), "test9.parquet")
    data = [
        {"x": []},
        {"x": [{"y": {"q": one}}]},
        {"x": [{"y": {"q": one}}, {"y": {"q": two}}, {"y": {"q": three}}]},
    ]
    ak.to_parquet(ak.Array(data), filename)
    array = ak.from_parquet(filename, lazy=True, lazy_cache_key="tmp")
    assert set(array.caches[0].keys()) == set()
    array.layout.field("x").array
    assert set(array.caches[0].keys()) == set(["tmp:off:x.list.item.y.q:x[0]"])
    assert np.asarray(array.layout.field("x").array.offsets).tolist() == [0, 0, 1, 4]
    assert set(array.caches[0].keys()) == set(["tmp:off:x.list.item.y.q:x[0]"])
    array.layout.field("x").array.content.field("y").array
    assert set(array.caches[0].keys()) == set(["tmp:off:x.list.item.y.q:x[0]"])
    array.layout.field("x").array.content.field("y").array.field("q").array
    assert set(array.caches[0].keys()) == set(
        ["tmp:off:x.list.item.y.q:x[0]", "tmp:col:x.list.item.y.q[0]"]
    )
    assert array.tolist() == data


@pytest.mark.parametrize("one,two,three", [(1, 2, 3), ("one", "two", "three")])
def test_10(one, two, three, tmp_path):
    filename = os.path.join(str(tmp_path), "test10.parquet")
    data = [
        {"x": []},
        {"x": [{"y": {"q": one}, "z": 1.1}]},
        {
            "x": [
                {"y": {"q": one}, "z": 1.1},
                {"y": {"q": two}, "z": 2.2},
                {"y": {"q": three}, "z": 3.3},
            ]
        },
    ]
    ak.to_parquet(ak.Array(data), filename)
    array = ak.from_parquet(filename, lazy=True, lazy_cache_key="tmp")
    assert set(array.caches[0].keys()) == set()
    array.layout.field("x").array
    assert set(array.caches[0].keys()) == set(["tmp:off:x.list.item.y.q:x[0]"])
    assert np.asarray(array.layout.field("x").array.offsets).tolist() == [0, 0, 1, 4]
    assert set(array.caches[0].keys()) == set(["tmp:off:x.list.item.y.q:x[0]"])
    array.layout.field("x").array.content.field("y").array
    assert set(array.caches[0].keys()) == set(["tmp:off:x.list.item.y.q:x[0]"])
    array.layout.field("x").array.content.field("y").array.field("q").array
    assert set(array.caches[0].keys()) == set(
        ["tmp:off:x.list.item.y.q:x[0]", "tmp:col:x.list.item.y.q[0]"]
    )
    array.layout.field("x").array.content.field("z").array
    assert set(array.caches[0].keys()) == set(
        [
            "tmp:off:x.list.item.y.q:x[0]",
            "tmp:col:x.list.item.y.q[0]",
            "tmp:col:x.list.item.z[0]",
        ]
    )
    assert array.tolist() == data
    array = ak.from_parquet(filename, lazy=True, lazy_cache_key="tmp")
    assert set(array.caches[0].keys()) == set()
    array.layout.field("x").array
    assert set(array.caches[0].keys()) == set(["tmp:off:x.list.item.y.q:x[0]"])
    assert np.asarray(array.layout.field("x").array.offsets).tolist() == [0, 0, 1, 4]
    assert set(array.caches[0].keys()) == set(["tmp:off:x.list.item.y.q:x[0]"])
    array.layout.field("x").array.content.field("y").array
    assert set(array.caches[0].keys()) == set(["tmp:off:x.list.item.y.q:x[0]"])
    array.layout.field("x").array.content.field("z").array
    assert set(array.caches[0].keys()) == set(
        ["tmp:off:x.list.item.y.q:x[0]", "tmp:col:x.list.item.z[0]"]
    )
    array.layout.field("x").array.content.field("y").array.field("q").array
    assert set(array.caches[0].keys()) == set(
        [
            "tmp:off:x.list.item.y.q:x[0]",
            "tmp:col:x.list.item.y.q[0]",
            "tmp:col:x.list.item.z[0]",
        ]
    )
    assert array.tolist() == data


@pytest.mark.parametrize("one,two,three", [(1, 2, 3), ("one", "two", "three")])
def test_11(one, two, three, tmp_path):
    filename = os.path.join(str(tmp_path), "test11.parquet")
    data = [
        {"x": []},
        {"x": [{"z": 1.1, "y": {"q": one}}]},
        {
            "x": [
                {"z": 1.1, "y": {"q": one}},
                {"z": 2.2, "y": {"q": two}},
                {"z": 3.3, "y": {"q": three}},
            ]
        },
    ]
    ak.to_parquet(ak.Array(data), filename)
    array = ak.from_parquet(filename, lazy=True, lazy_cache_key="tmp")
    assert set(array.caches[0].keys()) == set()
    array.layout.field("x").array
    assert set(array.caches[0].keys()) == set(["tmp:off:x.list.item.z:x[0]"])
    assert np.asarray(array.layout.field("x").array.offsets).tolist() == [0, 0, 1, 4]
    assert set(array.caches[0].keys()) == set(["tmp:off:x.list.item.z:x[0]"])
    array.layout.field("x").array.content.field("y").array
    assert set(array.caches[0].keys()) == set(["tmp:off:x.list.item.z:x[0]"])
    array.layout.field("x").array.content.field("y").array.field("q").array
    assert set(array.caches[0].keys()) == set(
        ["tmp:off:x.list.item.z:x[0]", "tmp:col:x.list.item.y.q[0]"]
    )
    array.layout.field("x").array.content.field("z").array
    assert set(array.caches[0].keys()) == set(
        [
            "tmp:off:x.list.item.z:x[0]",
            "tmp:col:x.list.item.y.q[0]",
            "tmp:col:x.list.item.z[0]",
        ]
    )
    assert array.tolist() == data
    array = ak.from_parquet(filename, lazy=True, lazy_cache_key="tmp")
    assert set(array.caches[0].keys()) == set()
    array.layout.field("x").array
    assert set(array.caches[0].keys()) == set(["tmp:off:x.list.item.z:x[0]"])
    assert np.asarray(array.layout.field("x").array.offsets).tolist() == [0, 0, 1, 4]
    assert set(array.caches[0].keys()) == set(["tmp:off:x.list.item.z:x[0]"])
    array.layout.field("x").array.content.field("y").array
    assert set(array.caches[0].keys()) == set(["tmp:off:x.list.item.z:x[0]"])
    array.layout.field("x").array.content.field("z").array
    assert set(array.caches[0].keys()) == set(
        ["tmp:off:x.list.item.z:x[0]", "tmp:col:x.list.item.z[0]"]
    )
    array.layout.field("x").array.content.field("y").array.field("q").array
    assert set(array.caches[0].keys()) == set(
        [
            "tmp:off:x.list.item.z:x[0]",
            "tmp:col:x.list.item.y.q[0]",
            "tmp:col:x.list.item.z[0]",
        ]
    )
    assert array.tolist() == data


@pytest.mark.parametrize("one,two,three", [(1, 2, 3), ("one", "two", "three")])
def test_12(one, two, three, tmp_path):
    filename = os.path.join(str(tmp_path), "test12.parquet")
    data = [
        {"x": {"y": []}},
        {"x": {"y": [[one]]}},
        {"x": {"y": [[one, two], [], [three]]}},
    ]
    ak.to_parquet(ak.Array(data), filename)
    array = ak.from_parquet(filename, lazy=True, lazy_cache_key="tmp")
    assert set(array.caches[0].keys()) == set()
    array.layout.field("x").array
    assert set(array.caches[0].keys()) == set()
    array.layout.field("x").array.field("y").array
    assert set(array.caches[0].keys()) == set(["tmp:lst:x.y[0]"])
    assert array.tolist() == data


@pytest.mark.parametrize("one,two,three", [(1, 2, 3), ("one", "two", "three")])
def test_13(one, two, three, tmp_path):
    filename = os.path.join(str(tmp_path), "test13.parquet")
    data = [
        {"x": {"y": [], "z": 1.1}},
        {"x": {"y": [[one]], "z": 2.2}},
        {"x": {"y": [[one, two], [], [three]], "z": 3.3}},
    ]
    ak.to_parquet(ak.Array(data), filename)
    array = ak.from_parquet(filename, lazy=True, lazy_cache_key="tmp")
    assert set(array.caches[0].keys()) == set()
    array.layout.field("x").array
    assert set(array.caches[0].keys()) == set()
    array.layout.field("x").array.field("z").array
    assert set(array.caches[0].keys()) == set(["tmp:col:x.z[0]"])
    array.layout.field("x").array.field("y").array
    assert set(array.caches[0].keys()) == set(["tmp:col:x.z[0]", "tmp:lst:x.y[0]"])
    assert array.tolist() == data
    array = ak.from_parquet(filename, lazy=True, lazy_cache_key="tmp")
    assert set(array.caches[0].keys()) == set()
    array.layout.field("x").array
    assert set(array.caches[0].keys()) == set()
    array.layout.field("x").array.field("y").array
    assert set(array.caches[0].keys()) == set(["tmp:lst:x.y[0]"])
    array.layout.field("x").array.field("z").array
    assert set(array.caches[0].keys()) == set(["tmp:lst:x.y[0]", "tmp:col:x.z[0]"])
    assert array.tolist() == data


@pytest.mark.parametrize("one,two,three", [(1, 2, 3), ("one", "two", "three")])
def test_14(one, two, three, tmp_path):
    filename = os.path.join(str(tmp_path), "test6.parquet")
    data = [
        {"x": [{"y": [], "z": 1.1}]},
        {"x": []},
        {"x": [{"y": [one, two, three], "z": 3.3}]},
    ]
    ak.to_parquet(ak.Array(data), filename)
    array = ak.from_parquet(filename, lazy=True, lazy_cache_key="tmp")
    assert set(array.caches[0].keys()) == set()
    array.layout.field("x").array
    assert set(array.caches[0].keys()) == set(["tmp:off:x.list.item.y.list.item:x[0]"])
    array.layout.field("x").array.content.field("z").array
    assert set(array.caches[0].keys()) == set(
        ["tmp:off:x.list.item.y.list.item:x[0]", "tmp:col:x.list.item.z[0]"]
    )
    array.layout.field("x").array.content.field("y").array
    assert set(array.caches[0].keys()) == set(
        [
            "tmp:off:x.list.item.y.list.item:x[0]",
            "tmp:col:x.list.item.z[0]",
            "tmp:lst:x.list.item.y[0]",
        ]
    )
    assert array.tolist() == data
    array = ak.from_parquet(filename, lazy=True, lazy_cache_key="tmp")
    assert set(array.caches[0].keys()) == set()
    array.layout.field("x").array
    assert set(array.caches[0].keys()) == set(["tmp:off:x.list.item.y.list.item:x[0]"])
    array.layout.field("x").array.content.field("y").array
    assert set(array.caches[0].keys()) == set(
        ["tmp:off:x.list.item.y.list.item:x[0]", "tmp:lst:x.list.item.y[0]"]
    )
    array.layout.field("x").array.content.field("z").array
    assert set(array.caches[0].keys()) == set(
        [
            "tmp:off:x.list.item.y.list.item:x[0]",
            "tmp:lst:x.list.item.y[0]",
            "tmp:col:x.list.item.z[0]",
        ]
    )
    assert array.tolist() == data


@pytest.mark.parametrize("one,two,three", [(1, 2, 3), ("one", "two", "three")])
def test_15(one, two, three, tmp_path):
    filename = os.path.join(str(tmp_path), "test15.parquet")
    data = [one, two, three]
    ak.to_parquet(ak.Array(data), filename)
    array = ak.from_parquet(filename, lazy=True, lazy_cache_key="tmp")
    assert set(array.caches[0].keys()) == set()
    array.layout.array
    assert set(array.caches[0].keys()) == set(["tmp:col:[0]"])
    assert array.tolist() == data


@pytest.mark.parametrize("one,two,three", [(1, 2, 3), ("one", "two", "three")])
def test_16(one, two, three, tmp_path):
    filename = os.path.join(str(tmp_path), "test15.parquet")
    data = [[one, two], [], [three]]
    ak.to_parquet(ak.Array(data), filename)
    array = ak.from_parquet(filename, lazy=True, lazy_cache_key="tmp")
    assert set(array.caches[0].keys()) == set()
    assert np.asarray(array.layout.array.offsets).tolist() == [0, 2, 2, 3]
    assert set(array.caches[0].keys()) == set(["tmp:lst:[0]"])
    assert array.tolist() == data


@pytest.mark.parametrize("one,two,three", [(1, 2, 3), ("one", "two", "three")])
def test_17(one, two, three, tmp_path):
    filename = os.path.join(str(tmp_path), "test15.parquet")
    data = [[{"x": one}, {"x": two}], [], [{"x": three}]]
    ak.to_parquet(ak.Array(data), filename)
    array = ak.from_parquet(filename, lazy=True, lazy_cache_key="tmp")
    assert set(array.caches[0].keys()) == set()
    assert np.asarray(array.layout.array.offsets).tolist() == [0, 2, 2, 3]
    assert set(array.caches[0].keys()) == set(["tmp:off:.list.item.x:[0]"])
    array.layout.array.content.field("x").array
    assert set(array.caches[0].keys()) == set(
        ["tmp:off:.list.item.x:[0]", "tmp:col:.list.item.x[0]"]
    )
    assert array.tolist() == data



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    assert ak.flatten(ak.Array([[1, 2, 3], [], [4, 5]]), axis=0).tolist() == [
        [1, 2, 3],
        [],
        [4, 5],
    ]
    assert ak.flatten(ak.Array([1, 2, 3, 4, 5]), axis=0).tolist() == [1, 2, 3, 4, 5]
    assert ak.flatten(ak.Array([[1, 2, 3], [], [4, 5]]), axis=-2).tolist() == [
        [1, 2, 3],
        [],
        [4, 5],
    ]
    assert ak.flatten(ak.Array([1, 2, 3, 4, 5]), axis=-1).tolist() == [1, 2, 3, 4, 5]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import os

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


pytest.importorskip("pyarrow.parquet")


def test(tmp_path):
    filename = os.path.join(tmp_path, "test.parquet")
    dog = ak.from_iter([1, 2, 5])
    cat = ak.from_iter([4])
    pets = ak.zip({"dog": dog[np.newaxis], "cat": cat[np.newaxis]}, depth_limit=1)
    ak.to_parquet(pets, filename)
    assert ak.from_parquet(filename).tolist() == pets.tolist()



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401
from awkward._v2._typetracer import Interval

typetracer = ak._v2._typetracer.TypeTracer.instance()


def test_getitem_at():
    concrete = ak._v2.contents.NumpyArray(np.arange(2 * 3 * 5).reshape(2, 3, 5) * 0.1)
    abstract = ak._v2.contents.NumpyArray(concrete.to(typetracer))

    assert concrete.shape == (2, 3, 5)
    assert abstract.shape == (Interval.exact(2), 3, 5)
    assert abstract[0].shape == (Interval.exact(3), 5)
    assert abstract[0][0].shape == (Interval.exact(5),)
    assert abstract[0][0][0] == 0

    assert abstract.form == concrete.form
    assert abstract.form.type == concrete.form.type

    assert abstract[0].form == concrete[0].form
    assert abstract[0].form.type == concrete[0].form.type



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


numba = pytest.importorskip("numba")


def test():
    @numba.njit
    def f1(array):
        return array[2][0].x

    array = ak.virtual(
        lambda: ak.Array([[{"x": True}, {"x": False}], [], [{"x": True}]])
    )
    assert f1(array) is True



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import awkward as ak

np = ak.nplike.NumpyMetadata.instance()


class Identifier(object):
    _numrefs = 0

    @staticmethod
    def newref():
        out = Identifier._numrefs
        Identifier._numrefs += 1
        return out

    def __init__(self, ref, fieldloc, data):
        self._ref = ref
        self._fieldloc = fieldloc
        if not isinstance(fieldloc, dict) or not all(
            ak._util.isint(k) and ak._util.isstr(v) for k, v in fieldloc.items()
        ):
            raise TypeError("Identifier fieldloc must be a dict of int -> str")
        self._nplike = ak.nplike.of(data)

        self._data = self._nplike.asarray(data, order="C")
        if len(self._data.shape) != 2:
            raise TypeError("Identifier data must be 2-dimensional")
        if self._data.dtype not in (np.dtype(np.int32), np.dtype(np.int64)):
            raise TypeError("Identifier data must be int32, int64")

    @classmethod
    def zeros(cls, ref, fieldloc, length, width, nplike, dtype):
        return Identifier(ref, fieldloc, nplike.zeros((length, width), dtype=dtype))

    @classmethod
    def empty(cls, ref, fieldloc, length, width, nplike, dtype):
        return Identifier(ref, fieldloc, nplike.empty((length, width), dtype=dtype))

    @property
    def ref(self):
        return self._ref

    @property
    def fieldloc(self):
        return self._fieldloc

    @property
    def data(self):
        return self._data

    @property
    def nplike(self):
        return self._nplike

    @property
    def dtype(self):
        return self._data.dtype

    @property
    def shape(self):
        return self._data.shape

    def __len__(self):
        return len(self._data)

    def width(self):
        return self._data.shape[1]

    def to64(self):
        return Identifier(self._ref, self._fieldloc, self._data.astype(np.int64))

    def __getitem__(self, where):
        return self._data[where]

    def __copy__(self):
        return Identifier(self._ref, self._fieldloc, self._data.copy())

    def __repr__(self):
        return self._repr("", "", "")

    def _repr(self, indent, pre, post):
        out = [indent, pre, "<Identifier ref="]
        out.append(repr(str(self._ref)))
        out.append(" fieldloc=")
        out.append(repr(str(self._fieldloc)))
        out.append(" dtype=")
        out.append(repr(str(self.dtype)))
        out.append(" shape=")
        out.append(repr(str(self.shape)))

        arraystr_lines = self._nplike.array_str(
            self._data, max_line_width=max(80 - len(indent) - 4, 40)
        ).split("\n")
        if len(arraystr_lines) > 9:
            arraystr_lines = arraystr_lines[:4] + [" ..."] + arraystr_lines[-4:]
        out.append(">\n" + indent + "    ")
        out.append(("\n" + indent + "    ").join(arraystr_lines))
        out.append("\n" + indent + "</Identifier>")

        out.append(post)
        return "".join(out)

    def convert_to(self, nplike):
        return Identifier(self._ref, self._fieldloc, nplike.asarray(self._data))

    def referentially_equal(self, other):
        return (
            self._ref == other._ref
            and self._fieldloc == other._fieldloc
            and self._data.ctypes.data == other._data.ctypes.data
            and self._data.shape == other._data.shape
            and self._data.strides == other._data.strides
            and self._data.dtype == other._data.dtype
        )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from __future__ import absolute_import

import pytest  # noqa: F401
import numpy as np  # noqa: F401
import awkward as ak  # noqa: F401


def test():
    a = ak.Array([[1, 2, 3, 4], [5, 6, 7, 8]])
    assert ak.unflatten(a, [2, 2, 2, 2], axis=1).tolist() == [
        [[1, 2], [3, 4]],
        [[5, 6], [7, 8]],
    ]
    assert ak.unflatten(a, 2, axis=1).tolist() == [[[1, 2], [3, 4]], [[5, 6], [7, 8]]]
